// SPDX-License-Identifier: Apache-2.0
// Copyright (C) 2015-2022, Xilinx Inc - All rights reserved
// Copyright (C) 2022-2024 Advanced Micro Devices, Inc. All rights reserved.
#include "xrt/deprecated/xrt.h"
