// SPDX-License-Identifier: Apache-2.0
// Copyright (C) 2024 Advanced Micro Devices, Inc. All rights reserved.

// The content of this file has moved.
// This file is no longer exported and is used internally
// within XRT for legacy reasons
#include "xrt/detail/ert.h"
