// SPDX-License-Identifier: Apache-2.0
// Copyright (C) 2024 Advanced Micro Devices, Inc. All rights reserved.
#ifndef XRT_ERROR_CODE_H_
#define XRT_ERROR_CODE_H_

// The content of this file has moved.  This file is no longer
// exported but is used internally within XRT in files built by DKMS
#include "xrt/detail/xrt_error_code.h"

#endif
