// SPDX-License-Identifier: Apache-2.0
// Copyright (C) 2024 Advanced Micro Devices, Inc. All rights reserved.
#pragma message("please replace experimental/xrt_bo.h with xrt/xrt_bo.h")
#include "xrt/xrt_bo.h"
