// SPDX-License-Identifier: Apache-2.0
// Copyright (C) 2024 Advanced Micro Devices, Inc. All rights reserved.
#pragma message("please replace experimental/xclbin_util.h with xrt/experimental/xclbin_util.h")
#include "xrt/experimental/xclbin_util.h"
